from pathlib import Path
from PIL import Image
from typing import Union, BinaryIO

from .config import Configuration
from .decorators import anywhere, ensure_client
from .misc import prepare_upload_data
from .task_response import TaskResponse
from .chunkr_base import ChunkrBase

class Chunkr(ChunkrBase):
    """Chunkr API client that works in both sync and async contexts"""
    
    @anywhere()
    @ensure_client()
    async def upload(
        self,
        file: Union[str, Path, BinaryIO, Image.Image],
        config: Configuration = None,
    ) -> TaskResponse:
        task = await self.create_task(file, config)
        return await task.poll()

    @anywhere()
    @ensure_client()
    async def update(self, task_id: str, config: Configuration) -> TaskResponse:
        task = await self.update_task(task_id, config)
        return await task.poll()

    @anywhere()
    @ensure_client()
    async def create_task(
        self,
        file: Union[str, Path, BinaryIO, Image.Image],
        config: Configuration = None,
    ) -> TaskResponse:
        files = prepare_upload_data(file, config)
        r = await self._client.post(
            f"{self.url}/api/v1/task", files=files, headers=self._headers()
        )
        r.raise_for_status()
        return TaskResponse(**r.json()).with_client(self)

    @anywhere()
    @ensure_client()
    async def update_task(self, task_id: str, config: Configuration) -> TaskResponse:
        files = prepare_upload_data(None, config)
        r = await self._client.patch(
            f"{self.url}/api/v1/task/{task_id}",
            files=files,
            headers=self._headers(),
        )
        r.raise_for_status()
        return TaskResponse(**r.json()).with_client(self)

    @anywhere()
    @ensure_client()
    async def get_task(self, task_id: str) -> TaskResponse:
        r = await self._client.get(
            f"{self.url}/api/v1/task/{task_id}", headers=self._headers()
        )
        r.raise_for_status()
        return TaskResponse(**r.json()).with_client(self)

    @anywhere()
    @ensure_client()
    async def delete_task(self, task_id: str) -> None:
        r = await self._client.delete(
            f"{self.url}/api/v1/task/{task_id}", headers=self._headers()
        )
        r.raise_for_status()

    @ensure_client()
    @anywhere()
    async def cancel_task(self, task_id: str) -> None:
        r = await self._client.get(
            f"{self.url}/api/v1/task/{task_id}/cancel", headers=self._headers()
        )
        r.raise_for_status()

    @anywhere()
    async def close(self):
        if self._client:
            await self._client.aclose()
            self._client = None


import functools
import asyncio
import httpx
from typing import Callable, Any, TypeVar, Awaitable, Union, overload
try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

T = TypeVar('T')
P = ParamSpec('P')

_sync_loop = None

@overload
def anywhere() -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Union[Awaitable[T], T]]]: ...

def anywhere():
    """Decorator that allows an async function to run anywhere - sync or async context."""
    def decorator(async_func: Callable[P, Awaitable[T]]) -> Callable[P, Union[Awaitable[T], T]]:
        @functools.wraps(async_func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Union[Awaitable[T], T]:
            global _sync_loop
            try:
                loop = asyncio.get_running_loop()
                return async_func(*args, **kwargs)
            except RuntimeError:
                if _sync_loop is None:
                    _sync_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(_sync_loop)
                try:
                    return _sync_loop.run_until_complete(async_func(*args, **kwargs))
                finally:
                    asyncio.set_event_loop(None)
        return wrapper
    return decorator

def ensure_client() -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Decorator that ensures a valid httpx.AsyncClient exists before executing the method"""
    def decorator(async_func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @functools.wraps(async_func)
        async def wrapper(self: Any, *args: P.args, **kwargs: P.kwargs) -> T:
            if not self._client or self._client.is_closed:
                self._client = httpx.AsyncClient()
            return await async_func(self, *args, **kwargs)
        return wrapper
    return decorator

def require_task() -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Decorator that ensures task has required attributes and valid client before execution"""
    def decorator(async_func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @functools.wraps(async_func)
        async def wrapper(self: Any, *args: P.args, **kwargs: P.kwargs) -> T:
            if not self.task_url:
                raise ValueError("Task URL not found")
            if not self._client:
                raise ValueError("Client not found")
            if not self._client._client or self._client._client.is_closed:
                self._client._client = httpx.AsyncClient()
            return await async_func(self, *args, **kwargs)
        return wrapper
    return decorator


from .config import Configuration
from .task_response import TaskResponse
from .auth import HeadersMixin
from abc import abstractmethod
from dotenv import load_dotenv
import httpx
import io
import json
import os
from pathlib import Path
from PIL import Image
import requests
from typing import BinaryIO, Tuple, Union


class ChunkrBase(HeadersMixin):
    """Base class with shared functionality for Chunkr API clients."""

    def __init__(self, url: str = None, api_key: str = None):
        load_dotenv()
        self.url = url or os.getenv("CHUNKR_URL") or "https://api.chunkr.ai"
        self._api_key = api_key or os.getenv("CHUNKR_API_KEY")
        if not self._api_key:
            raise ValueError(
                "API key must be provided either directly, in .env file, or as CHUNKR_API_KEY environment variable. You can get an api key at: https://www.chunkr.ai"
            )

        self.url = self.url.rstrip("/")
        self._client = httpx.AsyncClient()

    def _prepare_file(
        self, file: Union[str, Path, BinaryIO, Image.Image]
    ) -> Tuple[str, BinaryIO]:
        """Convert various file types into a tuple of (filename, file-like object).

        Args:
            file: Input file, can be:
                - String or Path to a file
                - URL string starting with http:// or https://
                - Base64 string
                - Opened binary file (mode='rb')
                - PIL/Pillow Image object

        Returns:
            Tuple[str, BinaryIO]: (filename, file-like object) ready for upload

        Raises:
            FileNotFoundError: If the file path doesn't exist
            TypeError: If the file type is not supported
            ValueError: If the URL is invalid or unreachable
            ValueError: If the MIME type is unsupported
        """
        # Handle URLs
        if isinstance(file, str) and (
            file.startswith("http://") or file.startswith("https://")
        ):
            response = requests.get(file)
            response.raise_for_status()
            file_obj = io.BytesIO(response.content)
            filename = Path(file.split("/")[-1]).name or "downloaded_file"
            return filename, file_obj

        # Handle base64 strings
        if isinstance(file, str) and "," in file and ";base64," in file:
            try:
                # Split header and data
                header, base64_data = file.split(",", 1)
                import base64

                file_bytes = base64.b64decode(base64_data)
                file_obj = io.BytesIO(file_bytes)

                # Try to determine format from header
                format = "bin"
                mime_type = header.split(":")[-1].split(";")[0].lower()

                # Map MIME types to file extensions
                mime_to_ext = {
                    "application/pdf": "pdf",
                    "application/vnd.openxmlformats-officedocument.wordprocessingml.document": "docx",
                    "application/msword": "doc",
                    "application/vnd.openxmlformats-officedocument.presentationml.presentation": "pptx",
                    "application/vnd.ms-powerpoint": "ppt",
                    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": "xlsx",
                    "application/vnd.ms-excel": "xls",
                    "image/jpeg": "jpg",
                    "image/png": "png",
                    "image/jpg": "jpg",
                }

                if mime_type in mime_to_ext:
                    format = mime_to_ext[mime_type]
                else:
                    raise ValueError(f"Unsupported MIME type: {mime_type}")

                return f"file.{format}", file_obj
            except Exception as e:
                raise ValueError(f"Invalid base64 string: {str(e)}")

        # Handle file paths
        if isinstance(file, (str, Path)):
            path = Path(file).resolve()
            if not path.exists():
                raise FileNotFoundError(f"File not found: {file}")
            return path.name, open(path, "rb")

        # Handle PIL Images
        if isinstance(file, Image.Image):
            img_byte_arr = io.BytesIO()
            format = file.format or "PNG"
            file.save(img_byte_arr, format=format)
            img_byte_arr.seek(0)
            return f"image.{format.lower()}", img_byte_arr

        # Handle file-like objects
        if hasattr(file, "read") and hasattr(file, "seek"):
            # Try to get the filename from the file object if possible
            name = (
                getattr(file, "name", "document")
                if hasattr(file, "name")
                else "document"
            )
            return Path(name).name, file

        raise TypeError(f"Unsupported file type: {type(file)}")

    def _prepare_upload_data(
        self,
        file: Union[str, Path, BinaryIO, Image.Image],
        config: Configuration = None,
    ) -> Tuple[dict, dict]:
        """Prepare files and data dictionaries for upload.

        Args:
            file: The file to upload
            config: Optional configuration settings

        Returns:
            Tuple[dict, dict]: (files dict, data dict) ready for upload
        """
        filename, file_obj = self._prepare_file(file)
        files = {"file": (filename, file_obj)}
        data = {}

        if config:
            config_dict = config.model_dump(mode="json", exclude_none=True)
            for key, value in config_dict.items():
                if isinstance(value, dict):
                    files[key] = (None, json.dumps(value), "application/json")
                else:
                    data[key] = value

        return files, data

    @abstractmethod
    def upload(
        self,
        file: Union[str, Path, BinaryIO, Image.Image],
        config: Configuration = None,
    ) -> TaskResponse:
        """Upload a file and wait for processing to complete.

        Args:
            file: The file to upload.
            config: Configuration options for processing. Optional.

        Examples:
        ```python
        # Upload from file path
        await chunkr.upload("document.pdf")

        # Upload from opened file
        with open("document.pdf", "rb") as f:
            await chunkr.upload(f)

        # Upload from URL
        await chunkr.upload("https://example.com/document.pdf")

        # Upload from base64 string (must include MIME type header)
        await chunkr.upload("data:application/pdf;base64,JVBERi0...")

        # Upload an image
        from PIL import Image
        img = Image.open("photo.jpg")
        await chunkr.upload(img)
        ```
        Returns:
            TaskResponse: The completed task response
        """
        pass

    @abstractmethod
    def update(
        self, task_id: str, config: Configuration
    ) -> TaskResponse:
        """Update a task by its ID and wait for processing to complete.

        Args:
            task_id: The ID of the task to update
            config: Configuration options for processing. Optional.

        Returns:
            TaskResponse: The updated task response
        """
        pass

    @abstractmethod
    def create_task(
        self,
        file: Union[str, Path, BinaryIO, Image.Image],
        config: Configuration = None,
    ) -> TaskResponse:
        """Upload a file for processing and immediately return the task response. It will not wait for processing to complete. To wait for the full processing to complete, use `task.poll()`.

        Args:
            file: The file to upload.
            config: Configuration options for processing. Optional.

        Examples:
        ```
        # Upload from file path
        task = await chunkr.create_task("document.pdf")

        # Upload from opened file
        with open("document.pdf", "rb") as f:
            task = await chunkr.create_task(f)

        # Upload from URL
        task = await chunkr.create_task("https://example.com/document.pdf")

        # Upload from base64 string (must include MIME type header)
        task = await chunkr.create_task("data:application/pdf;base64,JVBERi0xLjcKCjEgMCBvYmo...")

        # Upload an image
        from PIL import Image
        img = Image.open("photo.jpg")
        task = await chunkr.create_task(img)

        # Wait for the task to complete - this can be done when needed
        await task.poll()
        ```
        """
        pass

    @abstractmethod
    def update_task(
        self, task_id: str, config: Configuration
    ) -> TaskResponse:
        """Update a task by its ID and immediately return the task response. It will not wait for processing to complete. To wait for the full processing to complete, use `task.poll()`.

        Args:
            task_id: The ID of the task to update
            config: Configuration options for processing. Optional.

        Returns:
            TaskResponse: The updated task response
        """
        pass

    @abstractmethod
    def get_task(self, task_id: str) -> TaskResponse:
        """Get a task response by its ID.

        Args:
            task_id: The ID of the task to get

        Returns:
            TaskResponse: The task response
        """
        pass

    @abstractmethod
    def delete_task(self, task_id: str) -> None:
        """Delete a task by its ID.

        Args:
            task_id: The ID of the task to delete
        """
        pass

    @abstractmethod
    def cancel_task(self, task_id: str) -> None:
        """Cancel a task by its ID.

        Args:
            task_id: The ID of the task to cancel
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Close the client connection. 
        This should be called when you're done using the client to properly clean up resources."""
        pass


from .config import Configuration
import io
import json
from pathlib import Path
from PIL import Image
import requests
from typing import Union, Tuple, BinaryIO, Optional

def prepare_file(file: Union[str, Path, BinaryIO, Image.Image]) -> Tuple[str, BinaryIO]:
    """Convert various file types into a tuple of (filename, file-like object)."""
    # Handle URLs
    if isinstance(file, str) and (
        file.startswith("http://") or file.startswith("https://")
    ):
        response = requests.get(file)
        response.raise_for_status()

        # Try to get filename from Content-Disposition header first
        filename = None
        content_disposition = response.headers.get("Content-Disposition")
        if content_disposition and "filename=" in content_disposition:
            filename = content_disposition.split("filename=")[-1].strip("\"'")

        # If no Content-Disposition, try to get clean filename from URL path
        if not filename:
            from urllib.parse import urlparse, unquote

            parsed_url = urlparse(file)
            path = unquote(parsed_url.path)
            filename = Path(path).name if path else None

        # Fallback to default name if we couldn't extract one
        filename = filename or "downloaded_file"

        # Sanitize filename: remove invalid characters and limit length
        import re

        filename = re.sub(
            r'[<>:"/\\|?*%]', "_", filename
        )  # Replace invalid chars with underscore
        filename = re.sub(r"\s+", "_", filename)  # Replace whitespace with underscore
        filename = filename.strip("._")  # Remove leading/trailing dots and underscores
        filename = filename[:255]  # Limit length to 255 characters

        file_obj = io.BytesIO(response.content)
        return filename, file_obj

    # Handle base64 strings
    if isinstance(file, str) and "," in file and ";base64," in file:
        try:
            # Split header and data
            header, base64_data = file.split(",", 1)
            import base64

            file_bytes = base64.b64decode(base64_data)
            file_obj = io.BytesIO(file_bytes)

            # Try to determine format from header
            format = "bin"
            mime_type = header.split(":")[-1].split(";")[0].lower()

            # Map MIME types to file extensions
            mime_to_ext = {
                "application/pdf": "pdf",
                "application/vnd.openxmlformats-officedocument.wordprocessingml.document": "docx",
                "application/msword": "doc",
                "application/vnd.openxmlformats-officedocument.presentationml.presentation": "pptx",
                "application/vnd.ms-powerpoint": "ppt",
                "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": "xlsx",
                "application/vnd.ms-excel": "xls",
                "image/jpeg": "jpg",
                "image/png": "png",
                "image/jpg": "jpg",
            }

            if mime_type in mime_to_ext:
                format = mime_to_ext[mime_type]
            else:
                raise ValueError(f"Unsupported MIME type: {mime_type}")

            return f"file.{format}", file_obj
        except Exception as e:
            raise ValueError(f"Invalid base64 string: {str(e)}")

    # Handle file paths
    if isinstance(file, (str, Path)):
        path = Path(file).resolve()
        if not path.exists():
            raise FileNotFoundError(f"File not found: {file}")
        return path.name, open(path, "rb")

    # Handle PIL Images
    if isinstance(file, Image.Image):
        img_byte_arr = io.BytesIO()
        format = file.format or "PNG"
        file.save(img_byte_arr, format=format)
        img_byte_arr.seek(0)
        return f"image.{format.lower()}", img_byte_arr

    # Handle file-like objects
    if hasattr(file, "read") and hasattr(file, "seek"):
        # Try to get the filename from the file object if possible
        name = (
            getattr(file, "name", "document") if hasattr(file, "name") else "document"
        )
        return Path(name).name, file

    raise TypeError(f"Unsupported file type: {type(file)}")


def prepare_upload_data(
    file: Optional[Union[str, Path, BinaryIO, Image.Image]] = None,
    config: Optional[Configuration] = None,
) -> dict:
    """Prepare files and data dictionaries for upload.

    Args:
        file: The file to upload
        config: Optional configuration settings

    Returns:
        dict: (files dict) ready for upload
    """
    files = {}
    if file:
        filename, file_obj = prepare_file(file)
        files = {"file": (filename, file_obj)}

    if config:
        config_dict = config.model_dump(mode="json", exclude_none=True)
        for key, value in config_dict.items():
            files[key] = (None, json.dumps(value), "application/json")

    return files



import pytest
from pathlib import Path
from PIL import Image
import asyncio

from chunkr_ai import Chunkr
from chunkr_ai.models import (
    Configuration,
    GenerationStrategy,
    GenerationConfig,
    OcrStrategy,
    Pipeline,
    SegmentationStrategy,
    SegmentProcessing,
    ChunkProcessing,
    TaskResponse,
)

@pytest.fixture
def sample_path():
    return Path("tests/files/test.pdf")

@pytest.fixture
def sample_image():
    return Image.open("tests/files/test.jpg")

@pytest.fixture
def client():
    client = Chunkr()
    yield client

@pytest.mark.asyncio
async def test_send_file_path(client, sample_path):
    response = await client.upload(sample_path)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_file_path_str(client, sample_path):
    response = await client.upload(str(sample_path))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_opened_file(client, sample_path):
    with open(sample_path, "rb") as f:
        response = await client.upload(f)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_send_pil_image(client, sample_image):
    response = await client.upload(sample_image)
    assert response.task_id is not None
    assert response.status == "Succeeded"

@pytest.mark.asyncio
async def test_ocr_auto(client, sample_path):
    response = await client.upload(sample_path, Configuration(ocr_strategy=OcrStrategy.AUTO))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_expires_in(client, sample_path):
    response = await client.upload(sample_path, Configuration(expires_in=10))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None
    await asyncio.sleep(10)
    try:
        task = await client.get_task(response.task_id)
        assert False
    except Exception as e:
        print(e)
        assert True

@pytest.mark.asyncio
async def test_chunk_processing(client, sample_path):
    response = await client.upload(
        sample_path,
        Configuration(chunk_processing=ChunkProcessing(target_length=1024)),
    )
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_segmentation_strategy_page(client, sample_path):
    response = await client.upload(
        sample_path, Configuration(segmentation_strategy=SegmentationStrategy.PAGE)
    )
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_page_llm_html(client, sample_path):
    response = await client.upload(
        sample_path,
        Configuration(
            segmentation_strategy=SegmentationStrategy.PAGE,
            segment_processing=SegmentProcessing(
                page=GenerationConfig(html=GenerationStrategy.LLM)
            ),
        ),
    )
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_page_llm(client, sample_path):
    configuration = Configuration(
        segmentation_strategy=SegmentationStrategy.PAGE,
        segment_processing=SegmentProcessing(
            page=GenerationConfig(
                html=GenerationStrategy.LLM, markdown=GenerationStrategy.LLM
            )
        ),
    )
    response = await client.upload(sample_path, configuration)
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_delete_task(client, sample_path):
    response = await client.upload(sample_path)
    await client.delete_task(response.task_id)
    with pytest.raises(Exception):
        await client.get_task(response.task_id)

@pytest.mark.asyncio
async def test_delete_task_direct(client, sample_path):
    task = await client.upload(sample_path)
    assert isinstance(task, TaskResponse)
    assert task.task_id is not None
    assert task.status == "Succeeded"
    assert task.output is not None
    await client.delete_task(task.task_id)
    with pytest.raises(Exception):
        await client.get_task(task.task_id)

@pytest.mark.asyncio
async def test_cancel_task(client, sample_path):
    response = await client.create_task(sample_path)
    assert response.status == "Starting"
    await client.cancel_task(response.task_id)
    assert (await client.get_task(response.task_id)).status == "Cancelled"

@pytest.mark.asyncio
async def test_cancel_task_direct(client, sample_path):
    task = await client.create_task(sample_path)
    assert isinstance(task, TaskResponse)
    assert task.status == "Starting"
    await task.cancel()
    assert task.status == "Cancelled"

@pytest.mark.asyncio
async def test_update_task(client, sample_path):
    original_config = Configuration(
        segmentation_strategy=SegmentationStrategy.LAYOUT_ANALYSIS,
    )
    new_config = Configuration(
        segmentation_strategy=SegmentationStrategy.PAGE,
    )
    task = await client.upload(sample_path, original_config)
    task = await client.update(task.task_id, new_config)
    assert task.status == "Succeeded"
    assert task.output is not None
    assert task.configuration.segmentation_strategy == SegmentationStrategy.PAGE

@pytest.mark.asyncio
async def test_update_task_direct(client, sample_path):
    original_config = Configuration(
        segmentation_strategy=SegmentationStrategy.LAYOUT_ANALYSIS,
    )
    new_config = Configuration(
        segmentation_strategy=SegmentationStrategy.PAGE,
    )
    task = await client.upload(sample_path, original_config)
    task = await task.update(new_config)
    assert task.status == "Succeeded"
    assert task.output is not None
    assert task.configuration.segmentation_strategy == SegmentationStrategy.PAGE

@pytest.mark.asyncio
async def test_pipeline_type(client, sample_path):
    response = await client.upload(sample_path, Configuration(pipeline=Pipeline.AZURE))
    assert response.task_id is not None
    assert response.status == "Succeeded"
    assert response.output is not None

@pytest.mark.asyncio
async def test_client_lifecycle(client, sample_path):
    response1 = await client.upload(sample_path)
    await client.close()
    response2 = await client.upload(sample_path)
    assert response1.task_id is not None
    assert response2.task_id is not None

@pytest.mark.asyncio
async def test_task_operations_after_client_close(client, sample_path):
    task = await client.create_task(sample_path)
    await client.close()
    result = await task.poll()
    assert result.status == "Succeeded"



from pydantic import BaseModel, Field, model_validator, ConfigDict
from enum import Enum
from typing import Optional, List, Dict, Union, Type

class GenerationStrategy(str, Enum):
    LLM = "LLM"
    AUTO = "Auto"

class CroppingStrategy(str, Enum):
    ALL = "All"
    AUTO = "Auto"

class GenerationConfig(BaseModel):
    html: Optional[GenerationStrategy] = None
    llm: Optional[str] = None
    markdown: Optional[GenerationStrategy] = None
    crop_image: Optional[CroppingStrategy] = None

class SegmentProcessing(BaseModel):
    model_config = ConfigDict(populate_by_name=True, alias_generator=str.title)

    title: Optional[GenerationConfig] = Field(default=None, alias="Title")
    section_header: Optional[GenerationConfig] = Field(
        default=None, alias="SectionHeader"
    )
    text: Optional[GenerationConfig] = Field(default=None, alias="Text")
    list_item: Optional[GenerationConfig] = Field(default=None, alias="ListItem")
    table: Optional[GenerationConfig] = Field(default=None, alias="Table")
    picture: Optional[GenerationConfig] = Field(default=None, alias="Picture")
    caption: Optional[GenerationConfig] = Field(default=None, alias="Caption")
    formula: Optional[GenerationConfig] = Field(default=None, alias="Formula")
    footnote: Optional[GenerationConfig] = Field(default=None, alias="Footnote")
    page_header: Optional[GenerationConfig] = Field(default=None, alias="PageHeader")
    page_footer: Optional[GenerationConfig] = Field(default=None, alias="PageFooter")
    page: Optional[GenerationConfig] = Field(default=None, alias="Page")

class ChunkProcessing(BaseModel):
    target_length: Optional[int] = None

class Property(BaseModel):
    name: str
    prop_type: str
    description: Optional[str] = None
    default: Optional[str] = None

class JsonSchema(BaseModel):
    title: str
    properties: List[Property]

class OcrStrategy(str, Enum):
    ALL = "All"
    AUTO = "Auto"

class SegmentationStrategy(str, Enum):
    LAYOUT_ANALYSIS = "LayoutAnalysis"
    PAGE = "Page"

class BoundingBox(BaseModel):
    left: float
    top: float
    width: float
    height: float

class OCRResult(BaseModel):
    bbox: BoundingBox
    text: str
    confidence: Optional[float]

class SegmentType(str, Enum):
    CAPTION = "Caption"
    FOOTNOTE = "Footnote"
    FORMULA = "Formula"
    LIST_ITEM = "ListItem"
    PAGE = "Page"
    PAGE_FOOTER = "PageFooter"
    PAGE_HEADER = "PageHeader"
    PICTURE = "Picture"
    SECTION_HEADER = "SectionHeader"
    TABLE = "Table"
    TEXT = "Text"
    TITLE = "Title"

class Segment(BaseModel):
    bbox: BoundingBox
    content: str
    page_height: float
    html: Optional[str] = None
    image: Optional[str] = None
    markdown: Optional[str] = None
    ocr: List[OCRResult]
    page_number: int
    page_width: float
    segment_id: str
    segment_type: SegmentType

class Chunk(BaseModel):
    chunk_id: str
    chunk_length: int
    segments: List[Segment]

class ExtractedJson(BaseModel):
    data: Dict

class OutputResponse(BaseModel):
    chunks: List[Chunk]
    file_name: Optional[str]
    page_count: Optional[int]
    pdf_url: Optional[str]

class Model(str, Enum):
    FAST = "Fast"
    HIGH_QUALITY = "HighQuality"

class Pipeline(str, Enum):
    AZURE = "Azure"

class Configuration(BaseModel):
    chunk_processing: Optional[ChunkProcessing] = None
    expires_in: Optional[int] = None
    high_resolution: Optional[bool] = None
    model: Optional[Model] = None
    ocr_strategy: Optional[OcrStrategy] = None
    segment_processing: Optional[SegmentProcessing] = None
    segmentation_strategy: Optional[SegmentationStrategy] = None
    pipeline: Optional[Pipeline] = None
    
class OutputConfiguration(Configuration):
    input_file_url: Optional[str] = None
    json_schema: Optional[Union[JsonSchema, Type[BaseModel], BaseModel]] = None
    
    @model_validator(mode="before")
    def map_deprecated_fields(cls, values: Dict) -> Dict:
        if isinstance(values, dict) and "target_chunk_length" in values:
            target_length = values.pop("target_chunk_length")
            if target_length is not None:
                values["chunk_processing"] = values.get("chunk_processing", {}) or {}
                values["chunk_processing"]["target_length"] = target_length
        return values
    
class Status(str, Enum):
    STARTING = "Starting"
    PROCESSING = "Processing"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELLED = "Cancelled"



from datetime import datetime
from typing import TypeVar, Optional, Generic
from pydantic import BaseModel, PrivateAttr
import asyncio

from .config import Configuration, OutputConfiguration, OutputResponse, Status
from .protocol import ChunkrClientProtocol
from .misc import prepare_upload_data
from .decorators import anywhere, require_task

T = TypeVar("T", bound="TaskResponse")

class TaskResponse(BaseModel, Generic[T]):
    configuration: OutputConfiguration
    created_at: datetime
    expires_at: Optional[datetime]
    finished_at: Optional[datetime]
    message: str
    output: Optional[OutputResponse]
    started_at: Optional[datetime]
    status: Status
    task_id: str
    task_url: Optional[str]
    _client: Optional[ChunkrClientProtocol] = PrivateAttr(default=None)

    def with_client(self, client: ChunkrClientProtocol) -> T:
        self._client = client
        return self

    def _check_status(self) -> Optional[T]:
        """Helper method to check task status and handle completion/failure"""
        if self.status == "Failed":
            raise ValueError(self.message)
        if self.status not in ("Starting", "Processing"):
            return self
        return None

    async def _poll_request(self) -> dict:
        try:
            if not self._client._client:
                raise ValueError("Client not found")
            r = await self._client._client.get(
                self.task_url, headers=self._client._headers()
            )
            r.raise_for_status()
            return r.json()
        except (ConnectionError, TimeoutError) as _:
            print("Connection error while polling the task, retrying...")
            await asyncio.sleep(0.5)
        except Exception:
            raise

    @anywhere()
    @require_task()
    async def poll(self) -> T:
        """Poll the task for completion."""
        while True:
            j = await self._poll_request()
            updated = TaskResponse(**j).with_client(self._client)
            self.__dict__.update(updated.__dict__)
            if res := self._check_status():
                return res
            await asyncio.sleep(0.5)

    @anywhere()
    @require_task()
    async def update(self, config: Configuration) -> T:
        """Update the task configuration."""
        f = prepare_upload_data(None, config)
        r = await self._client._client.patch(
            self.task_url, files=f, headers=self._client._headers()
        )
        r.raise_for_status()
        updated = TaskResponse(**r.json()).with_client(self._client)
        self.__dict__.update(updated.__dict__)
        return await self.poll()

    @anywhere()
    @require_task()
    async def delete(self) -> T:
        """Delete the task."""
        r = await self._client._client.delete(
            self.task_url, headers=self._client._headers()
        )
        r.raise_for_status()
        return self

    @anywhere()
    @require_task()
    async def cancel(self) -> T:
        """Cancel the task."""
        r = await self._client._client.get(
            f"{self.task_url}/cancel", headers=self._client._headers()
        )
        r.raise_for_status()
        return await self.poll()

    def html(self) -> str:
        """Get the full HTML of the task"""
        return self._get_content("html")

    def markdown(self) -> str:
        """Get the full markdown of the task"""
        return self._get_content("markdown")

    def content(self) -> str:
        """Get the full content of the task"""
        return self._get_content("content")

    def _get_content(self, t: str) -> str:
        if not self.output:
            return ""
        parts = []
        for c in self.output.chunks:
            for s in c.segments:
                v = getattr(s, t)
                if v:
                    parts.append(v)
        return "\n".join(parts)
